@@ -1,19 +1,19 @@
 /*  OpenVLBI - Open Source Very Long Baseline Interferometry
-*   Copyright © 2017-2022  Ilia Platone
+*   Copyright © 2017-2023  Ilia Platone
 *
-*   This program is free software; you can redistribute it and/or
-*   modify it under the terms of the GNU Lesser General Public
-*   License as published by the Free Software Foundation; either
-*   version 3 of the License, or (at your option) any later version.
+*   This program is free software; you can redistribute it and/or modify
+*   it under the terms of the GNU General Public License as published by
+*   the Free Software Foundation; either version 2 of the License, or
+*   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
-*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-*   Lesser General Public License for more details.
+*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+*   GNU General Public License for more details.
 *
-*   You should have received a copy of the GNU Lesser General Public License
-*   along with this program; if not, write to the Free Software Foundation,
-*   Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+*   You should have received a copy of the GNU General Public License along
+*   with this program; if not, write to the Free Software Foundation, Inc.,
+*   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include <stdio.h>
@@ -74,12 +74,13 @@ const char* vlbi_get_version()
     return VLBI_VERSION_STRING;
 }
 
-static double getDelay(double time, NodeCollection *nodes, VLBIBaseline *b, double Ra, double Dec,
+static double getDelay(double time, NodeCollection *nodes, VLBIBaseline *b, double Ra, double Dec, double Distance,
                        double wavelength)
 {
     double delay = 0.0;
     b->setRa(Ra);
     b->setDec(Dec);
+    b->setDistance(Distance);
     b->setRelative(nodes->isRelative());
     b->setWaveLength(wavelength);
     b->setTime(time);
@@ -88,25 +89,29 @@ static double getDelay(double time, NodeCollection *nodes, VLBIBaseline *b, doub
     return delay;
 }
 
-void vlbi_get_offsets(vlbi_context ctx, double J200Time, const char* node1, const char* node2, double Ra, double Dec,
-                      double *offset1,
-                      double *offset2)
+double vlbi_get_offset(vlbi_context ctx, double J2000Time, const char* node, double Ra, double Dec, double Distance)
 {
     NodeCollection* nodes = (NodeCollection*)ctx;
-    char baseline[150];
-    sprintf(baseline, "%s_%s", node1, node2);
-    VLBIBaseline *b = nodes->getBaselines()->Get(baseline);
-    if(b != nullptr) {
-        b->setRa(Ra);
-        b->setDec(Dec);
-        double max_delay = 0;
-        int farest = 0, x, y;
-        for (x = 0; x < nodes->Count(); x++)
+    char baseline[VLBI_BASELINE_NAME_SIZE];
+    BaselineCollection *collection = nodes->getBaselines();
+    double offset = 0.0;
+    collection->setRa(Ra);
+    collection->setDec(Dec);
+    collection->setDistance(Distance);
+    collection->setFrequency(1);
+    VLBIBaseline *bl = nullptr;
+    double max_delay = 0;
+    int farest = 0, x, y;
+    for (x = 0; x < nodes->count(); x++)
+    {
+        for (y = x + 1; y < nodes->count(); y++)
         {
-            for (y = x + 1; y < nodes->Count(); y++)
-            {
-                sprintf(baseline, "%s_%s", nodes->At(x)->getName(), nodes->At(y)->getName());
-                double delay = getDelay(J200Time, nodes, b, nodes->getBaselines()->Get(baseline)->getRa(), nodes->getBaselines()->Get(baseline)->getDec(), nodes->getBaselines()->Get(baseline)->getWaveLength());
+            sprintf(baseline, "%s_%s", nodes->at(x)->getName(), nodes->at(y)->getName());
+            if(!collection->contains(baseline))
+                sprintf(baseline, "%s_%s", nodes->at(y)->getName(), nodes->at(x)->getName());
+            if(collection->contains(baseline)) {
+                bl = collection->get(baseline);
+                double delay = getDelay(J2000Time, nodes, bl, bl->getRa(), bl->getDec(), bl->getDistance(), bl->getWaveLength());
                 if(fabs(delay) > max_delay)
                 {
                     max_delay = fabs(delay);
@@ -117,19 +122,16 @@ void vlbi_get_offsets(vlbi_context ctx, double J200Time, const char* node1, cons
                 }
             }
         }
-        BaselineCollection *collection = nodes->getBaselines();
-        VLBIBaseline *bl = nullptr;
-        sprintf(baseline, "%s_%s", b->getNode1()->getName(), nodes->At(farest)->getName());
-        if(collection->Contains(baseline)) {
-            bl = collection->Get(baseline);
-            *offset1 = getDelay(J200Time, nodes, bl, bl->getRa(), bl->getDec(), bl->getWaveLength());
-        }
-        sprintf(baseline, "%s_%s", b->getNode2()->getName(), nodes->At(farest)->getName());
-        if(collection->Contains(baseline)) {
-            bl = collection->Get(baseline);
-            *offset2 = getDelay(J200Time, nodes, bl, bl->getRa(), bl->getDec(), bl->getWaveLength());
-        }
     }
+    bl = nullptr;
+    sprintf(baseline, "%s_%s", node, nodes->at(farest)->getName());
+    if(!collection->contains(baseline))
+        sprintf(baseline, "%s_%s", nodes->at(farest)->getName(), node);
+    if(collection->contains(baseline)) {
+        bl = collection->get(baseline);
+        offset = getDelay(J2000Time, nodes, bl, bl->getRa(), bl->getDec(), bl->getDistance(), bl->getWaveLength());
+    }
+    return offset;
 }
 
 static void* fillplane(void *arg)
@@ -139,6 +141,8 @@ static void* fillplane(void *arg)
     {
         VLBIBaseline *b;
         NodeCollection *nodes;
+        NodeCollection *nodes_2nd;
+        BaselineCollection *baselines;
         bool moving_baseline;
         bool nodelay;
         int *stop;
@@ -153,7 +157,9 @@ static void* fillplane(void *arg)
     bool nodelay = argument->nodelay;
     NodeCollection *nodes = argument->nodes;
     if(nodes == nullptr)return nullptr;
-    BaselineCollection *baselines = nodes->getBaselines();
+    NodeCollection *nodes_2nd = argument->nodes_2nd;
+    if(nodes_2nd == nullptr)return nullptr;
+    BaselineCollection *baselines = argument->baselines;
     if(baselines == nullptr)return nullptr;
     dsp_stream_p parent = baselines->getStream();
     if(parent == nullptr)return nullptr;
@@ -167,8 +173,8 @@ static void* fillplane(void *arg)
     int e = 0;
     int s = 0;
     int i = 1;
-    double offset1;
-    double offset2;
+    int *pos = (int*)malloc(sizeof(int)*2);
+    double *offsets = (double*)malloc(sizeof(double)*baselines->getCorrelationOrder());
     int idx = 0;
     int oldidx = 0;
     int x;
@@ -177,25 +183,26 @@ static void* fillplane(void *arg)
     {
         if(*argument->stop)
             break;
-        for (x = 0; x < nodes->Count(); x++)
+        for (x = 0; x < nodes->count(); x++)
         {
             if(moving_baseline)
             {
-                nodes->At(x)->setLocation(l);
+                nodes->at(x)->setLocation(l);
             }
             else
             {
-                nodes->At(x)->setLocation(0);
+                nodes->at(x)->setLocation(0);
             }
         }
-        if(nodelay)
-        {
-            offset1 = 0.0;
-            offset2 = 0.0;
-        }
-        else
-        {
-            vlbi_get_offsets((void*)nodes, t, b->getNode1()->getName(), b->getNode2()->getName(), b->getRa(), b->getDec(), &offset1, &offset2);
+        for(int y = 0; y < baselines->getCorrelationOrder(); y++) {
+            if(nodelay)
+            {
+                offsets[y] = t;
+            }
+            else
+            {
+                offsets[y] = t + vlbi_get_offset((void*)nodes_2nd, t, b->getNode(y)->getName(), b->getRa(), b->getDec(), b->getDistance());
+            }
         }
         b->setTime(t);
         b->getProjection();
@@ -203,11 +210,13 @@ static void* fillplane(void *arg)
         int V = (int)b->getV() + v / 2;
         if(U >= 0 && U < u && V >= 0 && V < v)
         {
-            idx = (int)(U + V * u);
+            pos[0] = U;
+            pos[1] = V;
+            idx = dsp_stream_set_position(parent, pos);
             if(idx != oldidx)
             {
                 oldidx = idx;
-                val = b->Locked() ? b->Correlate(t) : b->Correlate(t + offset1, t + offset2);
+                val = b->Locked() ? b->Correlate(t) : b->Correlate(offsets);
                 if(mutex_initialized)
                 {
                     lock_mutex();
@@ -222,6 +231,8 @@ static void* fillplane(void *arg)
         s = l + 1;
         i = s - e;
     }
+    free(pos);
+    free(offsets);
     (*argument->nthreads)--;
     return nullptr;
 }
@@ -248,14 +259,13 @@ void vlbi_set_location(void *ctx, double lat, double lon, double el)
     nodes->stationLocation()->geographic.lon = lon;
     nodes->stationLocation()->geographic.el = el;
     nodes->setRelative(true);
-    nodes->getBaselines()->setRelative(true);
 }
 
 void vlbi_add_node(void *ctx, dsp_stream_p stream, const char *name, int geo)
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    nodes->Add(new VLBINode(stream, name, nodes->Count(), geo == 1));
+    nodes->add(new VLBINode(stream, name, nodes->count(), geo == 1));
 }
 
 void vlbi_copy_node(void *ctx, const char *name, const char *node)
@@ -263,8 +273,8 @@ void vlbi_copy_node(void *ctx, const char *name, const char *node)
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     if(vlbi_has_node(ctx, node)) {
-        VLBINode *n = nodes->Get(node);
-        nodes->Add(new VLBINode(dsp_stream_copy(n->getStream()), name, nodes->Count(), n->GeographicCoordinates()));
+        VLBINode *n = nodes->get(node);
+        nodes->add(new VLBINode(dsp_stream_copy(n->getStream()), name, nodes->count(), n->GeographicCoordinates()));
     }
 }
 
@@ -272,33 +282,33 @@ dsp_stream_p vlbi_get_node(void *ctx, const char *name)
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    return nodes->Get(name)->getStream();
+    return nodes->get(name)->getStream();
 }
 
 int vlbi_has_node(void *ctx, const char *name)
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    return nodes->Contains(name);
+    return nodes->contains(name);
 }
 
 int vlbi_get_nodes(void *ctx, vlbi_node** output)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    if(nodes->Count() > 0)
+    if(nodes->count() > 0)
     {
-        vlbi_node* out = (vlbi_node*)malloc(sizeof(vlbi_node) * (size_t)nodes->Count());
-        for(int x = 0; x < nodes->Count(); x++)
+        vlbi_node* out = (vlbi_node*)malloc(sizeof(vlbi_node) * (size_t)nodes->count());
+        for(int x = 0; x < nodes->count(); x++)
         {
-            out[x].GeographicLocation = nodes->At(x)->getGeographicLocation();
-            out[x].Location = nodes->At(x)->getLocation();
-            out[x].Geo = nodes->At(x)->GeographicCoordinates();
-            out[x].Stream = nodes->At(x)->getStream();
-            strcpy(out[x].Name, nodes->At(x)->getName());
-            out[x].Index = nodes->At(x)->getIndex();
+            out[x].GeographicLocation = nodes->at(x)->getGeographicLocation();
+            out[x].Location = nodes->at(x)->getLocation();
+            out[x].Geo = nodes->at(x)->GeographicCoordinates();
+            out[x].Stream = nodes->at(x)->getStream();
+            strcpy(out[x].Name, nodes->at(x)->getName());
+            out[x].Index = nodes->at(x)->getIndex();
         }
         *output = out;
-        return nodes->Count();
+        return nodes->count();
     }
     return 0;
 }
@@ -306,10 +316,10 @@ int vlbi_get_nodes(void *ctx, vlbi_node** output)
 void vlbi_del_node(void *ctx, const char *name)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    VLBINode* node = nodes->Get(name);
+    VLBINode* node = nodes->get(name);
     if(node != nullptr)
     {
-        nodes->Remove(name);
+        nodes->remove(name);
         node->~VLBINode();
     }
 }
@@ -318,7 +328,7 @@ void vlbi_filter_lp_node(void *ctx, const char *name, const char *node, double r
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    VLBINode *n = nodes->Get(node);
+    VLBINode *n = nodes->get(node);
     dsp_stream_p stream = dsp_stream_copy(n->getStream());
     dsp_fourier_dft(stream, 1);
     dsp_filter_lowpass(stream, radians);
@@ -333,7 +343,7 @@ void vlbi_filter_hp_node(void *ctx, const char *name, const char *node, double r
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    VLBINode *n = nodes->Get(node);
+    VLBINode *n = nodes->get(node);
     dsp_stream_p stream = dsp_stream_copy(n->getStream());
     dsp_fourier_dft(stream, 1);
     dsp_filter_highpass(stream, radians);
@@ -348,7 +358,7 @@ void vlbi_filter_bp_node(void *ctx, const char *name, const char *node, double l
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    VLBINode *n = nodes->Get(node);
+    VLBINode *n = nodes->get(node);
     dsp_stream_p stream = dsp_stream_copy(n->getStream());
     dsp_fourier_dft(stream, 1);
     dsp_filter_bandpass(stream, lo_radians, hi_radians);
@@ -363,7 +373,7 @@ void vlbi_filter_br_node(void *ctx, const char *name, const char *node, double l
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    VLBINode *n = nodes->Get(node);
+    VLBINode *n = nodes->get(node);
     dsp_stream_p stream = dsp_stream_copy(n->getStream());
     dsp_fourier_dft(stream, 1);
     dsp_filter_bandreject(stream, lo_radians, hi_radians);
@@ -378,7 +388,7 @@ void vlbi_add_model(void *ctx, dsp_stream_p stream, const char *name)
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    nodes->getModels()->Add(stream, name);
+    nodes->getModels()->add(stream, name);
 }
 
 void vlbi_copy_model(void *ctx, const char *name, const char *model)
@@ -394,15 +404,15 @@ int vlbi_get_models(void *ctx, dsp_stream_p** output)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     ModelCollection *models = nodes->getModels();
-    if(models->Count() > 0)
+    if(models->count() > 0)
     {
-        dsp_stream_p* out = (dsp_stream_p*)malloc(sizeof(dsp_stream_p) * models->Count());
-        for(int x = 0; x < models->Count(); x++)
+        dsp_stream_p* out = (dsp_stream_p*)malloc(sizeof(dsp_stream_p) * models->count());
+        for(int x = 0; x < models->count(); x++)
         {
-            out[x] = models->At(x);
+            out[x] = models->at(x);
         }
         *output = out;
-        return models->Count();
+        return models->count();
     }
     return 0;
 }
@@ -410,7 +420,7 @@ int vlbi_get_models(void *ctx, dsp_stream_p** output)
 dsp_stream_p vlbi_get_model(void *ctx, const char *name)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    dsp_stream_p model = nodes->getModels()->Get(name);
+    dsp_stream_p model = nodes->getModels()->get(name);
     if(model != nullptr)
     {
         dsp_buffer_stretch(model->buf, model->len, 0.0, dsp_t_max);
@@ -424,8 +434,8 @@ void vlbi_del_model(void *ctx, const char *name)
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     if(vlbi_has_model(ctx, name))
     {
-        dsp_stream_p model = nodes->getModels()->Get(name);
-        nodes->getModels()->Remove(name);
+        dsp_stream_p model = nodes->getModels()->get(name);
+        nodes->getModels()->remove(name);
         if(model != nullptr) {
             dsp_stream_free_buffer(model);
             dsp_stream_free(model);
@@ -433,88 +443,144 @@ void vlbi_del_model(void *ctx, const char *name)
     }
 }
 
+void vlbi_set_correlation_order(void *ctx, int order)
+{
+    NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
+    nodes->setCorrelationOrder(order);
+}
+
 int vlbi_get_baselines(void *ctx, vlbi_baseline** output)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     BaselineCollection* baselines = nodes->getBaselines();
-    if(baselines->Count() > 0)
+    if(baselines->count() > 0)
     {
-        vlbi_baseline* out = (vlbi_baseline*)malloc(sizeof(vlbi_baseline) * baselines->Count());
-        for(int x = 0; x < baselines->Count(); x++)
+        vlbi_baseline* out = (vlbi_baseline*)malloc(sizeof(vlbi_baseline) * baselines->count());
+        for(int x = 0; x < baselines->count(); x++)
         {
-            out[x].relative = baselines->At(x)->isRelative();
-            out[x].locked = baselines->At(x)->Locked();
-            out[x].Target = baselines->At(x)->getTarget();
-            out[x].Ra = baselines->At(x)->getRa();
-            out[x].Dec = baselines->At(x)->getDec();
-            out[x].baseline = baselines->At(x)->getBaseline();
-            out[x].u = baselines->At(x)->getU();
-            out[x].v = baselines->At(x)->getV();
-            out[x].delay = baselines->At(x)->getDelay();
-            out[x].WaveLength = baselines->At(x)->getWaveLength();
-            out[x].SampleRate = baselines->At(x)->getSampleRate();
-            out[x].Node1.GeographicLocation = baselines->At(x)->getNode1()->getGeographicLocation();
-            out[x].Node1.Location = baselines->At(x)->getNode1()->getLocation();
-            out[x].Node1.Geo = baselines->At(x)->getNode1()->GeographicCoordinates();
-            out[x].Node1.Stream = baselines->At(x)->getNode1()->getStream();
-            strcpy(out[x].Node1.Name, baselines->At(x)->getNode1()->getName());
-            out[x].Node1.Index = baselines->At(x)->getNode1()->getIndex();
-            out[x].Node2.GeographicLocation = baselines->At(x)->getNode2()->getGeographicLocation();
-            out[x].Node2.Location = baselines->At(x)->getNode2()->getLocation();
-            out[x].Node2.Geo = baselines->At(x)->getNode2()->GeographicCoordinates();
-            out[x].Node2.Stream = baselines->At(x)->getNode2()->getStream();
-            strcpy(out[x].Node2.Name, baselines->At(x)->getNode2()->getName());
-            out[x].Node2.Index = baselines->At(x)->getNode2()->getIndex();
-            strcpy(out[x].Name, baselines->At(x)->getName());
-            out[x].Stream = baselines->At(x)->getStream();
+            out[x].nodes_count = baselines->getCorrelationOrder();
+            out[x].Nodes = (vlbi_node*)malloc(sizeof(vlbi_node)*out[x].nodes_count);
+            out[x].relative = baselines->at(x)->isRelative();
+            out[x].locked = baselines->at(x)->Locked();
+            out[x].Target = baselines->at(x)->getTarget();
+            out[x].Ra = baselines->at(x)->getRa();
+            out[x].Dec = baselines->at(x)->getDec();
+            out[x].baseline = baselines->at(x)->getBaseline();
+            out[x].u = baselines->at(x)->getU();
+            out[x].v = baselines->at(x)->getV();
+            out[x].delay = baselines->at(x)->getDelay();
+            out[x].WaveLength = baselines->at(x)->getWaveLength();
+            out[x].SampleRate = baselines->at(x)->getSampleRate();
+            for(int y = 0; y < out[x].nodes_count; y++) {
+                out[x].Nodes[y].GeographicLocation = baselines->at(x)->getNode(y)->getGeographicLocation();
+                out[x].Nodes[y].Location = baselines->at(x)->getNode(y)->getLocation();
+                out[x].Nodes[y].Geo = baselines->at(x)->getNode(y)->GeographicCoordinates();
+                out[x].Nodes[y].Stream = baselines->at(x)->getNode(y)->getStream();
+                strcpy(out[x].Nodes[y].Name, baselines->at(x)->getNode(y)->getName());
+                out[x].Nodes[y].Index = baselines->at(x)->getNode(y)->getIndex();
+            }
+            strcpy(out[x].Name, baselines->at(x)->getName());
+            out[x].Stream = baselines->at(x)->getStream();
         }
         *output = out;
-        return baselines->Count();
+        return baselines->count();
     }
     return 0;
 }
 
-void vlbi_set_baseline_buffer(void *ctx, const char *node1, const char *node2, complex_t *buffer, int len)
+void vlbi_free_baselines(vlbi_baseline** baselines, int num_baselines)
+{
+    for(int x = 0; x < num_baselines; x++) {
+        free(baselines[x]->Nodes);
+    }
+    free(baselines);
+}
+
+void vlbi_set_baseline_buffer(void *ctx, const char **names, complex_t *buffer, int len)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    char name[150];
-    sprintf(name, "%s_%s", node1, node2);
-    VLBIBaseline *b = nodes->getBaselines()->Get(name);
-    if(b == nullptr) return;
+    VLBIBaseline *b = nullptr;
+    BaselineCollection *collection = nodes->getBaselines();
+    int x = 0;
+    for(x = 0; x < collection->count(); x++) {
+        b = collection->at(x);
+        int match = 0;
+        for(int y = 0; y < collection->getCorrelationOrder(); y++) {
+            for(int z = 0; z < collection->getCorrelationOrder(); z++) {
+                if(!strcmp(b->getNode(y)->getName(), names[z])) match ++;
+            }
+        }
+        if(match < collection->getCorrelationOrder()) continue;
+        break;
+    }
+    if(b == nullptr || x == collection->count()) return;
     dsp_stream_set_dim(b->getStream(), 0, len);
     dsp_stream_alloc_buffer(b->getStream(), len);
     b->getStream()->dft.pairs = buffer;
     b->Lock();
 }
 
-void vlbi_set_baseline_stream(void *ctx, const char *node1, const char *node2, dsp_stream_p stream)
+void vlbi_set_baseline_stream(void *ctx, const char **names, dsp_stream_p stream)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    char name[150];
-    sprintf(name, "%s_%s", node1, node2);
-    VLBIBaseline *b = nodes->getBaselines()->Get(name);
-    if(b == nullptr) return;
+    VLBIBaseline *b = nullptr;
+    BaselineCollection *collection = nodes->getBaselines();
+    int x = 0;
+    for(x = 0; x < collection->count(); x++) {
+        b = collection->at(x);
+        int match = 0;
+        for(int y = 0; y < collection->getCorrelationOrder(); y++) {
+            for(int z = 0; z < collection->getCorrelationOrder(); z++) {
+                if(!strcmp(b->getNode(y)->getName(), names[z])) match ++;
+            }
+        }
+        if(match < collection->getCorrelationOrder()) continue;
+        break;
+    }
+    if(b == nullptr || x == collection->count()) return;
     b->setStream(stream);
     b->Lock();
 }
 
-dsp_stream_p vlbi_get_baseline_stream(void *ctx, const char *node1, const char *node2)
+dsp_stream_p vlbi_get_baseline_stream(void *ctx, const char **names)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    char name[150];
-    sprintf(name, "%s_%s", node1, node2);
-    VLBIBaseline *b = nodes->getBaselines()->Get(name);
-    if(b == nullptr) return nullptr;
+    VLBIBaseline *b = nullptr;
+    BaselineCollection *collection = nodes->getBaselines();
+    int x = 0;
+    for(x = 0; x < collection->count(); x++) {
+        b = collection->at(x);
+        int match = 0;
+        for(int y = 0; y < collection->getCorrelationOrder(); y++) {
+            for(int z = 0; z < collection->getCorrelationOrder(); z++) {
+                if(!strcmp(b->getNode(y)->getName(), names[z])) match ++;
+            }
+        }
+        if(match < collection->getCorrelationOrder()) continue;
+        break;
+    }
+    if(b == nullptr || x == collection->count()) return nullptr;
     return b->getStream();
 }
 
-void vlbi_unlock_baseline(void *ctx, const char *node1, const char *node2)
+void vlbi_unlock_baseline(void *ctx, const char **names)
 {
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    char name[150];
-    sprintf(name, "%s_%s", node1, node2);
-    VLBIBaseline *b = nodes->getBaselines()->Get(name);
-    if(b == nullptr) return;
+    VLBIBaseline *b = nullptr;
+    BaselineCollection *collection = nodes->getBaselines();
+    int x = 0;
+    for(x = 0; x < collection->count(); x++) {
+        b = collection->at(x);
+        int match = 0;
+        for(int y = 0; y < collection->getCorrelationOrder(); y++) {
+            for(int z = 0; z < collection->getCorrelationOrder(); z++) {
+                if(!strcmp(b->getNode(y)->getName(), names[z])) match ++;
+            }
+        }
+        if(match < collection->getCorrelationOrder()) continue;
+        break;
+    }
+    if(b == nullptr || x == collection->count()) return;
     b->Unlock();
 }
 
@@ -526,40 +592,49 @@ void vlbi_get_uv_plot(vlbi_context ctx, const char *name, int u, int v, double *
     if(nodes == nullptr)return;
     BaselineCollection *baselines = nodes->getBaselines();
     if(baselines == nullptr)return;
+    NodeCollection *nodes_2nd = new NodeCollection();
+    for(int x = 0; x < nodes->count(); x++)
+        nodes_2nd->add(nodes->at(x));
+    nodes_2nd->setCorrelationOrder(2);
     int stop = 0;
     dsp_stream_p parent = baselines->getStream();
     dsp_buffer_set(parent->buf, parent->len, 0.0);
     baselines->setWidth(u);
     baselines->setHeight(v);
-    baselines->SetFrequency(freq);
-    baselines->SetSampleRate(sr);
+    baselines->setFrequency(freq);
+    baselines->setSampleRate(sr);
     baselines->setRa(target[0]);
     baselines->setDec(target[1]);
+    baselines->setDistance(target[2]);
     parent->child_count = 0;
-    pgarb("%ld nodes, %ld baselines\n", nodes->Count(), baselines->Count());
-    baselines->SetDelegate(delegate);
+    pgarb("%ld nodes, %ld baselines\n", nodes->count(), baselines->count());
+    baselines->setDelegate(delegate);
     pthread_attr_t attr;
     pthread_attr_init(&attr);
     pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
-    pthread_t *threads = (pthread_t*)malloc(sizeof(pthread_t) * baselines->Count());
+    pthread_t *threads = (pthread_t*)malloc(sizeof(pthread_t) * baselines->count());
     int threads_running = 0;
     int max_threads = (int)vlbi_max_threads(0);
     struct args
     {
         VLBIBaseline *b;
         NodeCollection *nodes;
+        NodeCollection *nodes_2nd;
+        BaselineCollection *baselines;
         bool moving_baseline;
         bool nodelay;
         int *stop;
         int *nthreads;
     };
-    args *argument = (args*)malloc(sizeof(args) * (size_t)baselines->Count());
-    for(int i = 0; i < baselines->Count(); i++)
+    args *argument = (args*)malloc(sizeof(args) * (size_t)baselines->count());
+    for(int i = 0; i < baselines->count(); i++)
     {
-        VLBIBaseline *b = baselines->At(i);
+        VLBIBaseline *b = baselines->at(i);
         if(b == nullptr)continue;
         argument[i].b = b;
         argument[i].nodes = nodes;
+        argument[i].nodes_2nd = nodes_2nd;
+        argument[i].baselines = baselines;
         argument[i].moving_baseline = moving_baseline;
         argument[i].nodelay = nodelay;
         argument[i].nthreads = &threads_running;
@@ -572,7 +647,7 @@ void vlbi_get_uv_plot(vlbi_context ctx, const char *name, int u, int v, double *
         threads_running++;
         pthread_create(&threads[i], &attr, fillplane, &argument[i]);
     }
-    for(int i = 0; i < baselines->Count(); i++)
+    for(int i = 0; i < baselines->count(); i++)
         pthread_join(threads[i], nullptr);
     free(argument);
     free(threads);
@@ -597,31 +672,24 @@ void vlbi_get_ifft(vlbi_context ctx, const char *name, const char *magnitude, co
         return;
     dsp_stream_p mag = vlbi_get_model(ctx, magnitude);
     dsp_stream_p phi = vlbi_get_model(ctx, phase);
-    int d = 0;
-    if(mag->dims == phi->dims)
-        for (d = 0; d < mag->dims && mag->sizes[d] == phi->sizes[d]; )
-            d++;
-    if(mag->dims == d)
-    {
-        dsp_stream_p ifft = nullptr;
-        if(vlbi_has_model(ctx, name))
-            ifft = vlbi_get_model(ctx, name);
-        else
-            ifft = dsp_stream_copy(mag);
-        ifft->phase = dsp_stream_copy(phi);
-        ifft->magnitude = dsp_stream_copy(mag);
-        dsp_buffer_stretch(ifft->phase->buf, ifft->phase->len, 0, PI * 2.0);
-        dsp_buffer_stretch(ifft->magnitude->buf, ifft->magnitude->len, 0, dsp_t_max);
-        dsp_buffer_set(ifft->buf, ifft->len, 0.0);
-        ifft->buf[0] = dsp_t_max;
-        dsp_fourier_idft(ifft);
-        dsp_stream_free_buffer(ifft->phase);
-        dsp_stream_free(ifft->phase);
-        dsp_stream_free_buffer(ifft->magnitude);
-        dsp_stream_free(ifft->magnitude);
-        if(!vlbi_has_model(ctx, name))
-            vlbi_add_model(ctx, ifft, name);
-    }
+    dsp_stream_p ifft = nullptr;
+    if(vlbi_has_model(ctx, name))
+        ifft = vlbi_get_model(ctx, name);
+    else
+        ifft = dsp_stream_copy(mag);
+    ifft->phase = dsp_stream_copy(phi);
+    ifft->magnitude = dsp_stream_copy(mag);
+    dsp_buffer_stretch(ifft->phase->buf, ifft->phase->len, 0, PI * 2.0);
+    dsp_buffer_stretch(ifft->magnitude->buf, ifft->magnitude->len, 0, dsp_t_max);
+    dsp_buffer_set(ifft->buf, ifft->len, 0.0);
+    ifft->buf[0] = dsp_t_max;
+    dsp_fourier_idft(ifft);
+    dsp_stream_free_buffer(ifft->phase);
+    dsp_stream_free(ifft->phase);
+    dsp_stream_free_buffer(ifft->magnitude);
+    dsp_stream_free(ifft->magnitude);
+    if(!vlbi_has_model(ctx, name))
+        vlbi_add_model(ctx, ifft, name);
 }
 
 void vlbi_get_fft(vlbi_context ctx, const char *name, const char *magnitude, const char *phase)
@@ -630,7 +698,7 @@ void vlbi_get_fft(vlbi_context ctx, const char *name, const char *magnitude, con
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     if(!vlbi_has_model(ctx, name))
         return;
-    dsp_stream_p fft = nodes->getModels()->Get(name);
+    dsp_stream_p fft = nodes->getModels()->get(name);
     fft->phase = nullptr;
     fft->magnitude = nullptr;
     dsp_fourier_dft(fft, 1);
@@ -646,21 +714,10 @@ void vlbi_apply_mask(vlbi_context ctx, const char *name, const char *stream, con
         return;
     if(!vlbi_has_model(ctx, mask))
         return;
-    dsp_stream_p masked = dsp_stream_copy(nodes->getModels()->Get(stream));
-    dsp_stream_p model = nodes->getModels()->Get(mask);
-    int d = 0;
-    if(masked->dims == model->dims)
-    {
-        for (d = 0; d < masked->dims && masked->sizes[d] == model->sizes[d]; )
-            d++;
-        if(masked->dims == d)
-        {
-            dsp_buffer_stretch(model->buf, model->len, 0.0, 1.0);
-            dsp_buffer_mul(masked, model->buf, model->len);
-            vlbi_add_model(ctx, masked, name);
-            return;
-        }
-    }
+    dsp_stream_p masked = dsp_stream_copy(nodes->getModels()->get(stream));
+    dsp_stream_p model = nodes->getModels()->get(mask);
+    dsp_stream_multiply(masked, model);
+    vlbi_add_model(ctx, masked, name);
 }
 
 void vlbi_apply_convolution_matrix(vlbi_context ctx, const char *name, const char *model, const char *matrix)
@@ -671,18 +728,18 @@ void vlbi_apply_convolution_matrix(vlbi_context ctx, const char *name, const cha
         return;
     if(!vlbi_has_model(ctx, matrix))
         return;
-    dsp_stream_p convoluted = dsp_stream_copy(nodes->getModels()->Get(model));
-    dsp_stream_p convolution = nodes->getModels()->Get(matrix);
-    if(convoluted->dims == convolution->dims)
+    dsp_stream_p convolved = dsp_stream_copy(nodes->getModels()->get(model));
+    dsp_stream_p convolution = nodes->getModels()->get(matrix);
+    if(convolved->dims == convolution->dims)
     {
-        dsp_fourier_dft(convoluted, 1);
+        dsp_fourier_dft(convolved, 1);
         dsp_fourier_dft(convolution, 1);
-        dsp_convolution_convolution(convoluted, convolution);
-        vlbi_add_model(ctx, convoluted, name);
-        dsp_stream_free_buffer(convoluted->magnitude);
-        dsp_stream_free(convoluted->magnitude);
-        dsp_stream_free_buffer(convoluted->phase);
-        dsp_stream_free(convoluted->phase);
+        dsp_convolution_convolution(convolved, convolution);
+        vlbi_add_model(ctx, convolved, name);
+        dsp_stream_free_buffer(convolved->magnitude);
+        dsp_stream_free(convolved->magnitude);
+        dsp_stream_free_buffer(convolved->phase);
+        dsp_stream_free(convolved->phase);
         dsp_stream_free_buffer(convolution->magnitude);
         dsp_stream_free(convolution->magnitude);
         dsp_stream_free_buffer(convolution->phase);
@@ -698,15 +755,13 @@ void vlbi_stack_models(vlbi_context ctx, const char *name, const char *model1, c
         return;
     if(!vlbi_has_model(ctx, model2))
         return;
-    dsp_stream_p stacked = dsp_stream_copy(nodes->getModels()->Get(model1));
-    dsp_stream_p model = nodes->getModels()->Get(model2);
-    if(stacked->dims == model->dims)
-    {
-        dsp_buffer_div1(stacked, 2);
-        dsp_buffer_div1(model, 2);
-        dsp_buffer_sum(stacked, model->buf, fmin(stacked->len, model->len));
-        vlbi_add_model(ctx, stacked, name);
-    }
+    dsp_stream_p stacked = dsp_stream_copy(nodes->getModels()->get(model1));
+    dsp_stream_p model = nodes->getModels()->get(model2);
+    dsp_t _min = dsp_stats_min(stacked->buf, stacked->len);
+    dsp_t _max = dsp_stats_max(stacked->buf, stacked->len);
+    dsp_stream_sum(stacked, model);
+    dsp_buffer_stretch(stacked->buf, stacked->len, _min, _max);
+    vlbi_add_model(ctx, stacked, name);
 }
 
 void vlbi_diff_models(vlbi_context ctx, const char *name, const char *model1, const char *model2)
@@ -717,8 +772,8 @@ void vlbi_diff_models(vlbi_context ctx, const char *name, const char *model1, co
         return;
     if(!vlbi_has_model(ctx, model2))
         return;
-    dsp_stream_p diff = dsp_stream_copy(nodes->getModels()->Get(model1));
-    dsp_stream_p model = nodes->getModels()->Get(model2);
+    dsp_stream_p diff = dsp_stream_copy(nodes->getModels()->get(model1));
+    dsp_stream_p model = nodes->getModels()->get(model2);
     dsp_t mn = dsp_stats_min(model->buf, model->len);
     dsp_t mx = dsp_stats_max(model->buf, model->len);
     if(diff->dims == model->dims)
@@ -734,7 +789,7 @@ void vlbi_shift(vlbi_context ctx, const char *name)
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     if(!vlbi_has_model(ctx, name))
         return;
-    dsp_stream_p shifted = nodes->getModels()->Get(name);
+    dsp_stream_p shifted = nodes->getModels()->get(name);
     dsp_buffer_shift(shifted);
 }
 
@@ -742,7 +797,7 @@ int vlbi_has_model(void *ctx, const char *name)
 {
     pfunc;
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
-    return nodes->getModels()->Contains(name);
+    return nodes->getModels()->contains(name);
 }
 
 void vlbi_add_model_from_png(void *ctx, char *filename, const char *name)
@@ -762,7 +817,7 @@ void vlbi_add_model_from_png(void *ctx, char *filename, const char *name)
                 vlbi_add_model(nodes, file[c], model);
             }
         }
-        vlbi_add_model(nodes, dsp_stream_copy(file[channels]), name);
+        vlbi_add_model(nodes, file[channels], name);
         free(file);
     }
 }
@@ -806,7 +861,7 @@ void vlbi_add_model_from_fits(void *ctx, char *filename, const char *name)
                 vlbi_add_model(nodes, file[c], model);
             }
         }
-        vlbi_add_model(nodes, dsp_stream_copy(file[channels]), name);
+        vlbi_add_model(nodes, file[channels], name);
         free(file);
     }
 }
@@ -820,10 +875,10 @@ void vlbi_get_model_to_png(void *ctx, char *filename, const char *name)
     char *modelname = (char*)malloc(strlen(name)+5);
     int c = 0;
     sprintf(modelname, "%s_%d", name, c);
-    dsp_stream_p model = nodes->getModels()->Get(modelname);
+    dsp_stream_p model = nodes->getModels()->get(modelname);
     if(model == nullptr) {
         file = (dsp_stream_p*)realloc(file, sizeof(dsp_stream_p)* (size_t)(components+1));
-        model = nodes->getModels()->Get(name);
+        model = nodes->getModels()->get(name);
         if(model != nullptr)
             file[components++] = model;
     }
@@ -831,7 +886,7 @@ void vlbi_get_model_to_png(void *ctx, char *filename, const char *name)
         file = (dsp_stream_p*)realloc(file, sizeof(dsp_stream_p)* (size_t)(components+1));
         sprintf(modelname, "%s_%d", name, components);
         file[components++] = model;
-        model = nodes->getModels()->Get(modelname);
+        model = nodes->getModels()->get(modelname);
     }
     if(components > 0)
         dsp_file_write_png_composite(filename, components, 9, file);
@@ -847,10 +902,10 @@ void vlbi_get_model_to_jpeg(void *ctx, char *filename, const char *name)
     char *modelname = (char*)malloc(strlen(name)+5);
     int c = 0;
     sprintf(modelname, "%s_%d", name, c);
-    dsp_stream_p model = nodes->getModels()->Get(modelname);
+    dsp_stream_p model = nodes->getModels()->get(modelname);
     if(model == nullptr) {
         file = (dsp_stream_p*)realloc(file, sizeof(dsp_stream_p)* (size_t)(components+1));
-        model = nodes->getModels()->Get(name);
+        model = nodes->getModels()->get(name);
         if(model != nullptr)
             file[components++] = model;
     }
@@ -858,7 +913,7 @@ void vlbi_get_model_to_jpeg(void *ctx, char *filename, const char *name)
         file = (dsp_stream_p*)realloc(file, sizeof(dsp_stream_p)* (size_t)(components+1));
         sprintf(modelname, "%s_%d", name, components);
         file[components++] = model;
-        model = nodes->getModels()->Get(modelname);
+        model = nodes->getModels()->get(modelname);
     }
     if(components > 0)
         dsp_file_write_jpeg_composite(filename, components, 90, file);
@@ -874,10 +929,10 @@ void vlbi_get_model_to_fits(void *ctx, char *filename, const char *name)
     char *modelname = (char*)malloc(strlen(name)+5);
     int c = 0;
     sprintf(modelname, "%s_%d", name, c);
-    dsp_stream_p model = nodes->getModels()->Get(modelname);
+    dsp_stream_p model = nodes->getModels()->get(modelname);
     if(model == nullptr) {
         file = (dsp_stream_p*)realloc(file, sizeof(dsp_stream_p)* (size_t)(components+1));
-        model = nodes->getModels()->Get(name);
+        model = nodes->getModels()->get(name);
         if(model != nullptr)
             file[components++] = model;
     }
@@ -885,7 +940,7 @@ void vlbi_get_model_to_fits(void *ctx, char *filename, const char *name)
         file = (dsp_stream_p*)realloc(file, sizeof(dsp_stream_p)* (size_t)(components+1));
         sprintf(modelname, "%s_%d", name, components);
         file[components++] = model;
-        model = nodes->getModels()->Get(modelname);
+        model = nodes->getModels()->get(modelname);
     }
     if(components > 0)
         dsp_file_write_fits_composite(filename, components, 16, file);
@@ -898,7 +953,7 @@ void vlbi_add_node_from_fits(void *ctx, char *filename, const char *name, int ge
     NodeCollection *nodes = (ctx != nullptr) ? (NodeCollection*)ctx : vlbi_nodes;
     dsp_stream_p stream = vlbi_file_read_fits(filename);
     if(stream != nullptr)
-        nodes->Add(new VLBINode(stream, name, nodes->Count(), geo == 1));
+        nodes->add(new VLBINode(stream, name, nodes->count(), geo == 1));
 }
 
 void vlbi_add_nodes_from_sdfits(void *ctx, char *filename, const char *name, int geo)
@@ -910,6 +965,6 @@ void vlbi_add_nodes_from_sdfits(void *ctx, char *filename, const char *name, int
     if(stream != nullptr)
     {
         for(int i = 0; i < n; i++)
-            nodes->Add(new VLBINode(stream[i], name, nodes->Count(), geo == 1));
+            nodes->add(new VLBINode(stream[i], name, nodes->count(), geo == 1));
     }
 }
