@@ -1,3 +1,21 @@
+/*  OpenVLBI - Open Source Very Long Baseline Interferometry
+*   Copyright © 2017-2023  Ilia Platone
+*
+*   This program is free software; you can redistribute it and/or modify
+*   it under the terms of the GNU General Public License as published by
+*   the Free Software Foundation; either version 2 of the License, or
+*   (at your option) any later version.
+*
+*   This program is distributed in the hope that it will be useful,
+*   but WITHOUT ANY WARRANTY; without even the implied warranty of
+*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+*   GNU General Public License for more details.
+*
+*   You should have received a copy of the GNU General Public License along
+*   with this program; if not, write to the Free Software Foundation, Inc.,
+*   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+*/
+
 #include "vlbi_server.h"
 #include <cstdio>
 #include <cstdlib>
@@ -7,6 +25,9 @@
 #include <signal.h>
 #include <vlbi/base64.h>
 #include <vlbi/instancecollection.h>
+#ifdef _WIN32
+#include "getline.c"
+#endif
 
 static int is_running = 1;
 
@@ -22,6 +43,14 @@ VLBI::Server::Server()
     Bps = 8;
     w = 128;
     h = 128;
+    if(nullptr == tmpdir)
+        tmpdir = getenv("TMPDIR");
+    if(nullptr == tmpdir)
+        tmpdir = getenv("TMP");
+    if(nullptr == tmpdir)
+        tmpdir = getenv("TEMP");
+    if(nullptr == tmpdir)
+        tmpdir = (char*)"/tmp";
     context = (char*)malloc(9);
     strcpy(context, "OpenVLBI\0");
     contexts = new InstanceCollection();
@@ -29,9 +58,9 @@ VLBI::Server::Server()
 
 VLBI::Server::~Server()
 {
-    if(GetContext() != nullptr)
+    if(getContext() != nullptr)
     {
-        vlbi_exit(GetContext());
+        vlbi_exit(getContext());
     }
 }
 
@@ -42,10 +71,11 @@ static double coverage_delegate(double x, double y)
     return 1.0;
 }
 
-void VLBI::Server::AddNode(const char *name, char *b64)
+void VLBI::Server::addNode(const char *name, char *b64)
 {
     char filename[128];
-    strcpy(filename, "tmp_nodeXXXXXX");
+    strcpy(filename, tmpdir);
+    strcat(filename, "/tmp_nodeXXXXXX");
     int fd = mkstemp(filename);
     if(fd > -1)
     {
@@ -56,15 +86,16 @@ void VLBI::Server::AddNode(const char *name, char *b64)
         (void)written;
         free(buf);
         close(fd);
-        vlbi_add_node_from_fits(GetContext(), filename, name, true);
+        vlbi_add_node_from_fits(getContext(), filename, name, true);
         unlink(filename);
     }
 }
 
-void VLBI::Server::AddNodes(const char *name, char *b64)
+void VLBI::Server::addNodes(const char *name, char *b64)
 {
     char filename[128];
-    strcpy(filename, "tmp_nodeXXXXXX");
+    strcpy(filename, tmpdir);
+    strcat(filename, "/tmp_nodeXXXXXX");
     int fd = mkstemp(filename);
     if(fd > -1)
     {
@@ -75,12 +106,12 @@ void VLBI::Server::AddNodes(const char *name, char *b64)
         (void)written;
         free(buf);
         close(fd);
-        vlbi_add_nodes_from_sdfits(GetContext(), filename, name, true);
+        vlbi_add_nodes_from_sdfits(getContext(), filename, name, true);
         unlink(filename);
     }
 }
 
-void VLBI::Server::AddNode(const char *name, dsp_location *locations, void *buf, int bytelen, timespec starttime, bool geo)
+void VLBI::Server::addNode(const char *name, dsp_location *locations, void *buf, int bytelen, timespec starttime, bool geo)
 {
     dsp_stream_p node = dsp_stream_new();
     int len = bytelen * 8 / abs(Bps);
@@ -111,87 +142,103 @@ void VLBI::Server::AddNode(const char *name, dsp_location *locations, void *buf,
     }
     node->location = locations;
     memcpy(&node->starttimeutc, &starttime, sizeof(timespec));
-    vlbi_add_node(GetContext(), node, name, geo);
+    vlbi_add_node(getContext(), node, name, geo);
 }
 
-void VLBI::Server::DelNode(const char *name)
+void VLBI::Server::delNode(const char *name)
 {
-    vlbi_del_node(GetContext(), name);
+    vlbi_del_node(getContext(), name);
 }
 
 void VLBI::Server::Plot(const char *name, int flags)
 {
-    double coords[3] = { Ra, Dec };
+    double coords[3] = { Ra, Dec, DBL_MAX };
     if((flags & plot_flags_custom_delegate) == 0) {
         setDelegate((flags & plot_flags_uv_coverage) != 0 ? coverage_delegate : vlbi_default_delegate);
     }
-    vlbi_get_uv_plot(GetContext(), name, w, h, coords, Freq, SampleRate, (flags & plot_flags_synced) != 0,
+    vlbi_get_uv_plot(getContext(), name, w, h, coords, Freq, SampleRate, (flags & plot_flags_synced) != 0,
                      (flags & plot_flags_moving_baseline) != 0,
                      getDelegate(), nullptr);
 }
 
 void VLBI::Server::Idft(const char *model, const char *magnitude, const char *phase)
 {
-    vlbi_get_ifft(GetContext(), model, magnitude, phase);
+    vlbi_get_ifft(getContext(), model, magnitude, phase);
 }
 
 void VLBI::Server::Dft(const char *model, const char *magnitude, const char *phase)
 {
-    vlbi_get_fft(GetContext(), model, magnitude, phase);
+    vlbi_get_fft(getContext(), model, magnitude, phase);
 }
 
 void VLBI::Server::Stack(const char *name, const char *model1, const char *model2)
 {
-    vlbi_stack_models(GetContext(), name, model1, model2);
+    vlbi_stack_models(getContext(), name, model1, model2);
 }
 
 void VLBI::Server::Diff(const char *name, const char *model1, const char *model2)
 {
-    vlbi_diff_models(GetContext(), name, model1, model2);
+    vlbi_diff_models(getContext(), name, model1, model2);
 }
 
-void VLBI::Server::Convolute(const char *name, const char *model1, const char *model2)
+void VLBI::Server::Convolve(const char *name, const char *model1, const char *model2)
 {
-    vlbi_apply_convolution_matrix(GetContext(), name, model1, model2);
+    vlbi_apply_convolution_matrix(getContext(), name, model1, model2);
 }
 
 void VLBI::Server::Mask(const char *name, const char *model, const char *mask)
 {
-    vlbi_apply_mask(GetContext(), name, model, mask);
+    vlbi_apply_mask(getContext(), name, model, mask);
+}
+
+void VLBI::Server::Copy(const char *name, const char *model)
+{
+    vlbi_copy_model(getContext(), name, model);
+}
+
+void VLBI::Server::CopyNode(const char *name, const char *node)
+{
+    vlbi_copy_model(getContext(), name, node);
 }
 
 void VLBI::Server::Shift(const char *name)
 {
-    vlbi_shift(GetContext(), name);
+    vlbi_shift(getContext(), name);
 }
 
 void VLBI::Server::LowPass(const char *name, const char *node, double freq)
 {
-    vlbi_filter_lp_node(GetContext(), name, node, freq);
+    vlbi_filter_lp_node(getContext(), name, node, freq);
 }
 
 void VLBI::Server::HighPass(const char *name, const char *node, double freq)
 {
-    vlbi_filter_hp_node(GetContext(), name, node, freq);
+    vlbi_filter_hp_node(getContext(), name, node, freq);
 }
 
 void VLBI::Server::BandPass(const char *name, const char *node, double lofreq, double hifreq)
 {
-    vlbi_filter_bp_node(GetContext(), name, node, lofreq, hifreq);
+    vlbi_filter_bp_node(getContext(), name, node, lofreq, hifreq);
 }
 
 void VLBI::Server::BandReject(const char *name, const char *node, double lofreq, double hifreq)
 {
-    vlbi_filter_br_node(GetContext(), name, node, lofreq, hifreq);
+    vlbi_filter_br_node(getContext(), name, node, lofreq, hifreq);
 }
 
-dsp_stream_p VLBI::Server::GetModel(const char *name)
+dsp_stream_p VLBI::Server::getModel(const char *name)
 {
-    return vlbi_get_model(GetContext(), name);
+    return vlbi_get_model(getContext(), name);
 }
 
-char* VLBI::Server::GetModel(const char *name, char *format)
+void VLBI::Server::setCorrelationOrder(int order)
 {
+    vlbi_set_correlation_order(getContext(), order);
+}
+
+char* VLBI::Server::getModel(const char *name, char *format)
+{
+    if(!vlbi_has_model(getContext(), name)) return nullptr;
     char filename[128];
     int channels = 1;
     int fd = -1;
@@ -199,44 +246,43 @@ char* VLBI::Server::GetModel(const char *name, char *format)
     ssize_t outlen = 0;
     unsigned char *buf = nullptr;
     unsigned char *b64 = nullptr;
-    strcpy(filename, "tmp_modelXXXXXX");
-    dsp_stream_p model = vlbi_get_model(GetContext(), name);
+    strcpy(filename, tmpdir);
+    strcat(filename, "/tmp_modelXXXXXX");
+    fd = mkstemp(filename);
+    dsp_stream_p model = vlbi_get_model(getContext(), name);
     dsp_stream_p *stream = (dsp_stream_p*)malloc(sizeof(dsp_stream_p) * (size_t)(channels + 1));
     for(int c = 0; c <= channels; c++)
-        stream[c] = model;
-    fd = mkstemp(filename);
+        stream[c] = dsp_stream_copy(model);
     if(fd > -1)
     {
-        close(fd);
         if(!strcmp(format, "jpeg"))
             dsp_file_write_jpeg_composite(filename, channels, 100, stream);
         if(!strcmp(format, "png"))
             dsp_file_write_png_composite(filename, channels, 9, stream);
         if(!strcmp(format, "fits"))
             dsp_file_write_fits_composite(filename, channels, 16, stream);
-        FILE* f = fopen(filename, "r");
+        FILE* f = fdopen(fd, "rb+");
         fseek(f, 0, SEEK_END);
         len = ftell(f);
         rewind(f);
         len -= ftell(f);
         buf = (unsigned char*)malloc((size_t)len);
-        size_t nread = fread(buf, (size_t)len, 1, f);
-        (void)nread;
-        fclose(f);
-        unlink(filename);
-        outlen = len * 4 / 3 + 4;
+        size_t nread = fread(buf, 1, (size_t)len, f);
+        outlen = nread * 4 / 3 + 4;
         b64 = (unsigned char*)malloc((size_t)outlen);
-        to64frombits(b64, buf, (int)len);
+        to64frombits(b64, buf, (int)nread);
         free(buf);
+        close(fd);
+        unlink(filename);
         return (char*)b64;
     }
     return nullptr;
 }
 
-int VLBI::Server::GetModels(char** names)
+int VLBI::Server::getModels(char** names)
 {
     dsp_stream_p* models;
-    int len = vlbi_get_models(GetContext(), &models);
+    int len = vlbi_get_models(getContext(), &models);
     if(len > 0)
     {
         names = (char**)malloc(sizeof(char*) * (size_t)len);
@@ -249,22 +295,24 @@ int VLBI::Server::GetModels(char** names)
     return 0;
 }
 
-void VLBI::Server::DelModel(const char *name)
+void VLBI::Server::delModel(const char *name)
 {
-    vlbi_del_model(GetContext(), name);
+    vlbi_del_model(getContext(), name);
 }
 
-void VLBI::Server::AddModel(const char *name, char *format, char *b64)
+void VLBI::Server::addModel(const char *name, char *format, char *b64)
 {
+    if(vlbi_has_model(getContext(), name)) return;
     char filename[128];
     int fd = -1;
     size_t b64len = 0;
     char *buf = nullptr;
-    strcpy(filename, "tmp_modelXXXXXX");
+    strcpy(filename, tmpdir);
+    strcat(filename, "/tmp_modelXXXXXX");
+    fd = mkstemp(filename);
     b64len = strlen(b64);
     if(b64len > 0)
     {
-        fd = mkstemp(filename);
         if(fd > -1)
         {
             buf = (char*)malloc(b64len * 3 / 4 + 4);
@@ -275,15 +323,15 @@ void VLBI::Server::AddModel(const char *name, char *format, char *b64)
             close(fd);
             if(!strcmp(format, "fits"))
             {
-                vlbi_add_model_from_fits(GetContext(), filename, name);
+                vlbi_add_model_from_fits(getContext(), filename, name);
             }
             if(!strcmp(format, "jpeg"))
             {
-                vlbi_add_model_from_jpeg(GetContext(), filename, name);
+                vlbi_add_model_from_jpeg(getContext(), filename, name);
             }
             if(!strcmp(format, "png"))
             {
-                vlbi_add_model_from_png(GetContext(), filename, name);
+                vlbi_add_model_from_png(getContext(), filename, name);
             }
             unlink(filename);
         }
@@ -297,10 +345,14 @@ void VLBI::Server::Parse()
     char *cmd = nullptr;
     char *arg = nullptr;
     char *value = nullptr;
-    char *str = nullptr;
-    ssize_t ofs = getdelim(&str, &len, (int)'\n', f);
-    (void)ofs;
-    if(str == nullptr)
+    char *str = (char*)malloc(1);
+    while(true) {
+        str[len] = 0;
+        fread(&str[len], 1, 1, f);
+        if(str[len++] == 0) break;
+        str = (char*)realloc(str, len + 8);
+    }
+    if(strlen(str) <= 0)
         return;
     if (len == 0)
         return;
@@ -325,7 +377,7 @@ void VLBI::Server::Parse()
         {
             if(!strcmp(arg, "context"))
             {
-                SetContext(value);
+                setContext(value);
             }
             else if(!strcmp(arg, "mask"))
             {
@@ -388,7 +440,7 @@ void VLBI::Server::Parse()
                 lon = (int)atof(t);
                 t = strtok(nullptr, ",");
                 el = (int)atof(t);
-                vlbi_set_location(GetContext(), lat, lon, el);
+                vlbi_set_location(getContext(), lat, lon, el);
             }
         }
         else if(!strcmp(cmd, "get"))
@@ -396,7 +448,7 @@ void VLBI::Server::Parse()
             if(!strcmp(arg, "models"))
             {
                 dsp_stream_p* models;
-                int n = vlbi_get_models(GetContext(), &models);
+                int n = vlbi_get_models(getContext(), &models);
                 for(int x = 0; x < n; x++)
                 {
                     fprintf(f, "Model #%d: name:%s width:%d height:%d\n", x,
@@ -407,7 +459,7 @@ void VLBI::Server::Parse()
             if(!strcmp(arg, "nodes"))
             {
                 vlbi_node* nodes;
-                int n = vlbi_get_nodes(GetContext(), &nodes);
+                int n = vlbi_get_nodes(getContext(), &nodes);
                 for(int x = 0; x < n; x++)
                 {
                     fprintf(f, "Node #%d: name:%s relative?:%s x:%lf y:%lf z:%lf latitude:%lf longitude:%lf elevation:%lf\n", nodes[x].Index,
@@ -419,7 +471,7 @@ void VLBI::Server::Parse()
             else if(!strcmp(arg, "baselines"))
             {
                 vlbi_baseline* baselines;
-                int n = vlbi_get_baselines(GetContext(), &baselines);
+                int n = vlbi_get_baselines(getContext(), &baselines);
                 for(int x = 0; x < n; x++)
                 {
                     fprintf(f, "Baseline #%d: name:%s samplerate:%lf wavelength:%lf custom?:%s\n", x, baselines[x].Name,
@@ -441,7 +493,7 @@ void VLBI::Server::Parse()
                 {
                     return;
                 }
-                char *base64 = GetModel(name, format);
+                char *base64 = getModel(name, format);
                 fwrite(base64, 1, strlen(base64), output);
                 free(base64);
             }
@@ -450,7 +502,7 @@ void VLBI::Server::Parse()
         {
             if(!strcmp(arg, "context"))
             {
-                AddContext(value);
+                addContext(value);
             }
             else if(!strcmp(arg, "node"))
             {
@@ -488,7 +540,7 @@ void VLBI::Server::Parse()
                 location.geographic.el = el;
                 if(len > 0 && geo > 0)
                 {
-                    AddNode(name, &location, buf, (int)len, vlbi_time_string_to_timespec(date), geo == 1);
+                    addNode(name, &location, buf, (int)len, vlbi_time_string_to_timespec(date), geo == 1);
                 }
             }
             else if(!strcmp(arg, "plot"))
@@ -602,53 +654,53 @@ void VLBI::Server::Parse()
                 {
                     return;
                 }
-                AddModel(name, format, b64);
+                addModel(name, format, b64);
             }
         }
         else if(!strcmp(cmd, "del"))
         {
             if(!strcmp(arg, "node"))
             {
-                DelNode(value);
+                delNode(value);
             }
             else if(!strcmp(arg, "context"))
             {
-                DelContext(value);
+                delContext(value);
             }
             else if(!strcmp(arg, "model"))
             {
-                DelModel(value);
+                delModel(value);
             }
         }
     }
 }
 
-void VLBI::Server::AddContext(const char *name)
+void VLBI::Server::addContext(const char *name)
 {
-    if(!contexts->Contains(name)) contexts->Add(vlbi_init(), name);
+    if(!contexts->contains(name)) contexts->add(vlbi_init(), name);
 }
 
-void VLBI::Server::SetContext(const char *name)
+void VLBI::Server::setContext(const char *name)
 {
-    if(contexts->Contains(name))
+    if(contexts->contains(name))
     {
         context = (char*)realloc(context, strlen(name));
         strcpy(context, name);
     }
 }
 
-vlbi_context VLBI::Server::GetContext()
+vlbi_context VLBI::Server::getContext()
 {
-    if(contexts->Contains(context)) return contexts->Get(context);
+    if(contexts->contains(context)) return contexts->get(context);
     return nullptr;
 }
 
-void VLBI::Server::DelContext(const char *name)
+void VLBI::Server::delContext(const char *name)
 {
-    if(contexts->Contains(name))
+    if(contexts->contains(name))
     {
-        vlbi_context ctx = contexts->Get(name);
-        contexts->Remove(name);
+        vlbi_context ctx = contexts->get(name);
+        contexts->remove(name);
         vlbi_exit(ctx);
     }
 }
@@ -661,13 +713,21 @@ static void sighandler(int signum)
     exit(0);
 }
 
-int main(int argc, char** argv)
+#ifdef _WIN32
+#include <windows.h>
+int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
 {
+    int argc = 0;
+    char **argv = {NULL};
+#else
+int main(int argc, char *argv[])
+{
+#endif
     int opt;
     dsp_set_app_name(argv[0]);
     dsp_set_stdout(stderr);
     dsp_set_stderr(stderr);
-    while ((opt = getopt(argc, argv, "t:f:o:v")) != -1)
+    while ((opt = getopt(argc, argv, "t:f:o:vh")) != -1)
     {
         switch (opt)
         {
@@ -675,32 +735,37 @@ int main(int argc, char** argv)
                 vlbi_max_threads((unsigned long)atol(optarg));
                 break;
             case 'f':
-                VLBI::server->SetInput(fopen (optarg, "rb+"));
+                VLBI::server->setInput(fopen (optarg, "rb+"));
                 break;
             case 'o':
-                VLBI::server->SetOutput(fopen (optarg, "a"));
+                VLBI::server->setOutput(fopen (optarg, "a"));
                 break;
             case 'v':
                 dsp_set_debug_level(dsp_get_debug_level()+1);
             break;
+            case 'h':
             default:
                 perr("Usage: %s [-t max_threads] [-f obs_file] [-o obs_file] [-v[v[v]]]\n", argv[0]);
                 exit(EXIT_FAILURE);
         }
     }
-    signal(SIGINT, sighandler);
-    signal(SIGKILL, sighandler);
+    signal(SIGTERM, sighandler);
     signal(SIGILL, sighandler);
+    signal(SIGINT, sighandler);
+#ifndef _WIN32
     signal(SIGSTOP, sighandler);
+    signal(SIGKILL, sighandler);
     signal(SIGQUIT, sighandler);
+#endif
     if(VLBI::server->Init(argc, argv))
     {
         while (is_running)
         {
-            if(feof(VLBI::server->GetInput()))
+            if(feof(VLBI::server->getInput()))
                 break;
             VLBI::server->Parse();
         }
     }
+    VLBI::server->~Server();
     return EXIT_SUCCESS;
 }
